#!/usr/bin/env perl

use strict;
use warnings;

use v5.16;
use feature qw<switch>;

use Config;
use Cwd qw<getcwd abs_path>;
use File::stat;

our $bitness;
our $owndir;
our $is_quickstream;
our $is_recording;
our $enable_lib_static;
our $update;

sub statsize {
	my $path = shift;
	return 0 if !defined($path);
	my $x = stat($path);
	return 0 if !defined($x);
	return $x->size;
}

sub get_bitness {
	given ($Config{archname}) {
		when (/^x86_64-.*/) { return 64; }
		# TODO: Test 32-bit
		when (/^i686-.*/)   { return 32; }
		default { die "Unsupported host"; }
	}
}

sub get_owndir {
	my $desired = abs_path("/mnt/main/tv/install");
	if (not defined $desired) {
		die "Cannot determine main path";
	}

	my $dir = abs_path(getcwd());
	if ($dir ne $desired) {
		die "Must run from ${desired}, not ${dir}";
	}
	if (not -d $dir) {
		die "Install directory is malformed";
	}

	return $dir;
}

sub check_disk_reserve {
	sub get_free_bytes {
		my $fs = shift;
		return `stat -f -c '%f' $fs` * `stat -f -c '%S' $fs`,
	}

	my $main_free = get_free_bytes("/mnt/main/");
	if ($main_free < 2**30) {
		die "Cannot update. Main partition does not have enough disk space. Clear some items from your library to remedy the situation.";
	}

	my $system_free = get_free_bytes("/");
	if ($system_free < 2**30) {
		my $log_size = statsize("/var/log/messages");
		my $reporting_log_size = statsize("/usr/share/castus/reporting/log");
		if ($system_free - $reporting_log_size < 2**30) {
			if ($system_free - $reporting_log_size - $log_size < 2**30) {
				die "Cannot update. System partition does not have enough disk space."
			}
			system("truncate", "-s", "0", "/usr/share/castus/reporting/log");
		}
		system("truncate", "-s", "0", "/var/log/messages");
	}
}

sub get_tar_metadata {
	my $tar = shift;
	my %meta;
	open(my $tarh, '-|', 'tar', '-xf', $tar, '.castus-update', '--to-stdout');
	while (<$tarh>) {
		chomp;
		if (/^([^=]+)=(.*)$/) {
			$meta{$1} = $2;
		}
	}
	close($tarh);
	return \%meta;
}

sub find_update_tarball {
	my @tarballs;

	if (opendir(my $dirh, getcwd())) {
		while (my $entry = readdir($dirh)) {
			next unless $entry;
			next unless $entry =~ /\.tar$/;
			my $meta = get_tar_metadata($entry);
			next unless exists $meta->{date};
			next unless $meta->{product} eq 'blackmagic-dv';
			if ($bitness == 64) {
				next unless $meta->{target} eq 'x86_64';
			} else { # TODO: Test 32-bit
				next unless $meta->{target} eq 'i686';
			}

			$meta->{path} = $entry;
			push(@tarballs, $meta);
		}
		closedir($dirh);
	} else {
		die "Unable to search for update package";
	}

	@tarballs = sort { $a->date cmp $b->date } @tarballs;
	if (+@tarballs == 0) {
		die "No suitable update tarball found";
	}
	return pop(@tarballs)
}

sub check_hash {
	my $file = shift;
	if (system("sha1sum", "-c", $file . ".sha1") != 0) {
		die "Update integrity check failed; try re-downloading";
	}
}

$bitness = get_bitness();
$owndir = get_owndir();
check_disk_reserve();
$update = find_update_tarball();
check_hash($update->{path});

# the tar metadata should list which kernels we have .ko files for.
# if our kernel is none of the above, let the user know, the update
# is less likely to work unless userspace matches kernelspace.
my $thiskernel = `uname -r`; chomp $thiskernel;
die unless $thiskernel ne "";

print "You are running kernel $thiskernel\n";

my $hasdriver = 0;
my $kverstr = "";
if (exists($update->{kernelversions})) {
	$kverstr = $update->{kernelversions};
}
for my $chkver (split(/ +/,$kverstr)) {
	if ($chkver eq $thiskernel) {
		$hasdriver = 1;
		last;
	}
}

print "NOTICE: After updating the Decklink software, it may be necessary to update\n";
print "        the Decklink card's firmware and then POWER OFF and then POWER ON\n";
print "        the Castus unit to fully reset the Decklink card firmware.\n";
print "\n";
print "        If a firmware update is not required, then the update can be done\n";
print "        without restarting your Castus unit.\n";
print "\n";

if ($hasdriver == 0) {
	print "WARNING: This update package does not have a driver for your current kernel ($thiskernel)\n";
	print "         It is possible to update the rest of the Decklink software but it is unlikely to\n";
	print "         work without updating the device driver for your system kernel as well.\n";
	print "\n";
	print "         Applying this update is not recommended.\n";
	print "\n";
}

print "Ready to apply the update? (Type Y or yes to confirm): "; $|++;
my $ready = <STDIN>; chomp $ready;
if ($ready !~ /^[yY]/) {
	print "Update aborted, exiting\n";
	exit 0;
}

# force the kernel driver to unload
print "\n";
print "Unloading kernel driver...\n";

# 12.x drivers need a helper process to run, shut it down too
system("/etc/init.d/service-decklink-helper stop >/dev/null 2>&1");

# trick: make it impossible for processes to open the Decklink device during this loop
#        by renaming /dev/blackmagic to /dev/blackmagic-UNLOAD
if ( -d "/dev/blackmagic" ) {
	system("rm -Rf /dev/blackmagic-UNLOAD");
	system("mv -n /dev/blackmagic /dev/blackmagic-UNLOAD");
}

# NTS: sysfs lists names with the underscore, not the hyphen.
my $ok = 0;
my @modules = ( 'blackmagic', 'blackmagic_io', 'snd_blackmagic', 'snd_blackmagic_io' );
# NTS: Nothing in Castus uses the snd_blackmagic_io driver, which is said to map the card as an ALSA output
my @load_modules = ( 'blackmagic', 'blackmagic_io' );

while ($ok == 0) {
	my $xx = system("fuser -k /dev/blackmagic/{dv,io}* /dev/blackmagic-UNLOAD/{dv,io}* >/dev/null 2>&1");
	if ( $xx != 0 ) {
		$ok = 1;
		my $rmmod = '';

		for my $module (@modules) {
			# use sysfs to detect if module is there
			if ( -d "/sys/module/$module" ) {
				$ok = 0;
				$rmmod .= " $module";
			}
		}

		if ($ok == 0) {
			print "...Attempting to remove: $rmmod\n";
			system("rmmod $rmmod 2>/dev/null");
		}
	}

	if ($ok == 0) {
		sleep 2;
	}
}

system("rm -Rf /dev/blackmagic-UNLOAD");
print "Kernel driver unloaded.\n";

# unpack the new software
print "Updating Decklink software and drivers...\n";
my $x = system("tar -xf ".$update->{path}." -C /");
print "WARNING: Errors occurred during unpacking.\n" if $x != 0;
print "Software update done\n";

# it's possible the kernel may not have known about the drivers yet
print "Updating kernel driver database...\n"; # it really is... we're not just quoting Windows 95 "updating driver database" here :)
system("depmod >/dev/null 2>&1");

# now load the new drivers
for my $module (@load_modules) {
	print "Loading kernel driver $module...\n";
	system("modprobe $module 2>/dev/null");
}

# wait
sleep 2;
print "Decklink update finished.\n";

# The 12.x SDK cannot access ANYTHING on the card until a helper process is started
my $modver = `modinfo blackmagic | grep '^version: '`; chomp $modver;
if ($modver ne "") {
	$modver =~ s/^version: +//;
}

if ($modver =~ m/^12\./) {
	system("/etc/init.d/service-decklink-helper start >/dev/null 2>&1");
	sleep 1;
}

# be nice and restart the user's auto-start outputs which often include SDI
system("c4mi_outctl -c start -o autostart");

# use the script we also installed to check firmware status
$x = system("blackmagic-dv-check-firmware-update");
# assume Linux semantics...
$x = ($x >> 8) & 0xFF;
if ($x == 1) {
	print "It is recommended that you update the card's firmware.\n";
	print "Begin firmware update? (Type Y or yes to confirm): "; $|++;
	my $ready2 = <STDIN>; chomp $ready2;
	if ($ready2 =~ m/^[yY]/) {
		print "\n";
		# this is where another script comes in handy
		system("/usr/sbin/blackmagic-dv-do-firmware-update");
	}
}
